/*
 *	Driver for the integrated watchdog in the AMD8111 core chipset.
 *
 *      Written 2003 by Andi Kleen, heavily based on Alan Cox's alim1535d_wdt.c
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License
 *	as published by the Free Software Foundation; version 2.
 */
 
#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/mm.h>
#include <linux/miscdevice.h>
#include <linux/watchdog.h>
#include <linux/reboot.h>
#include <linux/init.h>
#include <linux/pci.h>
#include <linux/ioport.h>
#include <asm/uaccess.h>
#include <asm/io.h>

static spinlock_t wdt8111_lock = SPIN_LOCK_UNLOCKED;

static unsigned char *wdog_base;

#ifdef CONFIG_WATCHDOG_NOWAYOUT
static int nowayout = 1;
#else
static int nowayout = 0;
#endif

MODULE_PARM(nowayout,"i");
MODULE_PARM_DESC(nowayout, "Watchdog cannot be stopped once started (default=CONFIG_WATCHDOG_NOWAYOUT)");

static unsigned long timer_alive;
static char wdt8111_expect_close;
static u32 wdt8111_timeout = 60;			/* 60 seconds */

static struct pci_dev *wdt8111_pci;

static void wdt8111_timer_start(struct pci_dev *pdev)
{
	u32 val;

	spin_lock(&wdt8111_lock);
	writel(wdt8111_timeout, wdog_base); 	/* set timeout */
	val = readl(wdog_base + 8);
	val &= ~(1<<2); /* reset the box on expire (otherwise power down) */
	val |= (1<<0) | (1<<7); /* enable / trigger */ 
	writel(val, wdog_base + 8);
	spin_unlock(&wdt8111_lock);
}

static void wdt8111_timer_stop (struct pci_dev *pdev)
{
	u32 val;

	spin_lock(&wdt8111_lock);
	val = readl(wdog_base + 8);
	val &= ~1;
	writel(val, wdog_base + 8);
	spin_unlock(&wdt8111_lock);
}

static int wdt8111_timer_settimer(struct pci_dev *pdev, unsigned long t)
{
	if (t >= 0xffff)
		return -EINVAL; 
	wdt8111_timeout = t;
	wdt8111_timer_start(pdev);
	return 0;
}

static int wdt8111_open(struct inode *inode, struct file *file)
{
	if (test_and_set_bit(0, &timer_alive))
		return -EBUSY;
	wdt8111_timer_start (wdt8111_pci);
	return 0;
}

static int wdt8111_release (struct inode *inode, struct file *file)
{
	/*
	 *      Shut off the timer.
	 */
	if (wdt8111_expect_close == 42 && !nowayout) {
		wdt8111_timer_stop(wdt8111_pci);
	} else {
		wdt8111_timer_start(wdt8111_pci);
		printk(KERN_CRIT  "amd8111_wdt: Unexpected close, not stopping watchdog!\n");
	}
	clear_bit(0, &timer_alive);
	wdt8111_expect_close = 0;
	return 0;
}

static ssize_t wdt8111_write (struct file *file, const char *data,
			      size_t len, loff_t * ppos)
{
	/*  Can't seek (pwrite) on this device  */
	if (ppos != &file->f_pos)
		return -ESPIPE;

	/* See if we got the magic character 'V' and reload the timer */
	if (len) {
		size_t i;

		wdt8111_expect_close = 0;

		/* scan to see wether or not we got the magic character */
		for (i = 0; i != len; i++) {
			u8 c;
			if(get_user(c, data+i))
				return -EFAULT;
			if (c == 'V')
				wdt8111_expect_close = 42;
		}

		/* someone wrote to us, we should reload the timer */
		wdt8111_timer_start(wdt8111_pci);
		return 1;
	}
	return 0;
}

static int wdt8111_ioctl (struct inode *inode, struct file *file,
			  unsigned int cmd, unsigned long arg)
{
	int options, retval = -EINVAL;
	u32 t;
	static struct watchdog_info ident = {
		options:		WDIOF_SETTIMEOUT | WDIOF_KEEPALIVEPING,
		firmware_version:	0,
		identity:		"AMD 8111 watchdog",
	};
	switch (cmd) {
		default:
			return -ENOTTY;
		case WDIOC_GETSUPPORT:
			if (copy_to_user((struct watchdog_info *) arg, &ident, sizeof (ident)))
				return -EFAULT;
			return 0;
		case WDIOC_GETSTATUS:
		case WDIOC_GETBOOTSTATUS:
			return put_user(0, (int *) arg);
		case WDIOC_SETOPTIONS:
			if (get_user (options, (int *) arg))
				return -EFAULT;
			if (options & WDIOS_DISABLECARD) {
				wdt8111_timer_stop(wdt8111_pci);
				retval = 0;
			}
			if (options & WDIOS_ENABLECARD) {
				wdt8111_timer_start(wdt8111_pci);
				retval = 0;
			}
			return retval;
		case WDIOC_KEEPALIVE:
			wdt8111_timer_start(wdt8111_pci);
			return 0;
		case WDIOC_SETTIMEOUT:
			if (get_user(t, (int *) arg))
				return -EFAULT;
			if (wdt8111_timer_settimer(wdt8111_pci, t))
			    return -EINVAL;
			wdt8111_timer_start(wdt8111_pci);
			/* Fall */
		case WDIOC_GETTIMEOUT:
			return put_user(wdt8111_timeout, (int *)arg);
	}
}

static struct pci_device_id wdt8111_pci_tbl[] __initdata = {
	{ PCI_VENDOR_ID_AMD, 0x7468, PCI_ANY_ID, PCI_ANY_ID,},
	{ 0, },
};
MODULE_DEVICE_TABLE (pci, wdt8111_pci_tbl);

static int __init wdt8111_find_watchdog(void)
{
	struct pci_dev *pdev;
	u32 wdog, tmp;
	unsigned long wdogmap = 0;
	
	/* Don't register a full driver because we share the device with 
	   others. */

	pdev = pci_find_device(PCI_VENDOR_ID_AMD, 0x7468, NULL);
	if(pdev == NULL) { 
		printk(KERN_ERR "amd8111_wdt: no device found.\n");
		return -ENODEV;
	}

	if(pci_enable_device(pdev)) {
		printk(KERN_ERR "amd8111_wdt: cannot enable device.\n");
		return -EIO;
	}
		
	wdt8111_pci = pdev;
	
	/*
	 *	Initialize the timer bits
	 */
	 
	pci_read_config_dword(pdev, 0xa8, &wdog);

	if (!(wdog & ~31)) { 
		printk(KERN_INFO "amd8111_wdt: watchdog base not set (%x)\n",wdog); 
		wdogmap = (unsigned long)kmalloc(64, GFP_DMA|GFP_KERNEL); 
		if (!wdogmap)
			return -ENOMEM;      
		wdog_base = (void *)((wdogmap + 31) & ~31UL);
		wdog |= __pa(wdog_base);		
	} else {
		wdog_base = __va(wdog & ~31);	/* checkme */
	}

	wdog &= ~((1<<2)|(1<<1)); 	/* disable silent mode and timer halt */
	wdog |= 1; /* enable */

	pci_write_config_dword(pdev, 0xa8, wdog); 

	pci_read_config_dword(pdev, 0xa8, &tmp); 
	if (tmp != wdog) { 
		/* Does not seem to work in some steppings. Handle gratefully. */
		printk(KERN_ERR "amd8111_wdt: cannot reread wdog base got %x expected %x\n", tmp, wdog);
		if (wdogmap) 
			kfree((void *)wdogmap); 
		return -EIO; 
	} 

	printk(KERN_INFO "amd8111_wdt: watchdog at %08lx.\n", __pa(wdog_base));

	return 0;
}

static struct file_operations wdt8111_fops = {
	owner:		THIS_MODULE,
	write:		wdt8111_write,
	ioctl:		wdt8111_ioctl,
	open:		wdt8111_open,
	release:	wdt8111_release,
};

static struct miscdevice wdt8111_miscdev = {
	minor:		WATCHDOG_MINOR,
	name:		"watchdog",
	fops:		&wdt8111_fops,
};

static int __init watchdog_init (void)
{
	if (!wdt8111_find_watchdog())
		return -ENODEV;
	if (misc_register (&wdt8111_miscdev) != 0) {
		printk (KERN_ERR "amd8111_wdt: cannot register watchdog device node.\n");
		return -EIO;
	}
	return 0;
}

static void __exit watchdog_cleanup (void)
{
	/* Don't kfree the allocated base. It stays around in the hardware
	   register until next time. */
	wdt8111_timer_stop(wdt8111_pci);	
	misc_deregister (&wdt8111_miscdev);
}

module_init(watchdog_init);
module_exit(watchdog_cleanup);

MODULE_AUTHOR("Andi Kleen, Alan Cox");
MODULE_DESCRIPTION("Driver for the AMD 8111 watchdog");
MODULE_LICENSE("GPL");
EXPORT_NO_SYMBOLS;
